# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

__title__ = "openc3"
__description__ = "Python Support for OpenC3 COSMOS"
__url__ = "https://github.com/OpenC3/cosmos"
__version__ = "6.4.1"
